﻿using UnityEngine;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Xml.Serialization;

/// <summary>
/// Implementation of the Occlusion Graph.
/// The occlusion graph is implemented using a hashmap that contains the direct occluder for each element in the scene.
/// In order to compute the occlusion graph, all objects are arranged into layers.
/// After computing an occlusion graph, it is saved to a xml file. If an occlusion Graph is already computed, it is loaded from a xml file.
/// </summary>
public class OcclusionGraph {

	private static float layerVisibilityRATIO= 0.9f;
	private static float graphOcclusionRATIO= 0.2f;

	public static string fileEnding = ".xml";

	private List<Layer> layers; 
	private Dictionary<string, List<Transform>> directOccluders = new Dictionary<string, List<Transform>> ();	//Saves for every element in the scene the direct occluders.

	/// <summary>
	/// Constructor of the Class <c>OcclusionGraph</c>
	/// </summary>
	public OcclusionGraph(){
		layers = new List<Layer> ();
	}
	
	/// <summary>
	/// Assignes all geometries in the scene to a layer. If the occlusion graph is already computed it is loaded from a xml file.
	/// </summary>
	/// <param name="camera">The camera at the moment</param>
	public void computeLayer(Camera camera){
		Debug.Log ("--------Compute Layer-----------");
		
		//TO get all geometries: GameObject.FindGameObjectsWithTag (Main.tagName);
		GameObject[] objectsInScene = GameObject.FindGameObjectsWithTag(Main.tagName);
		string fileName = "";
		List<string> objectNames = new List<string>();
		for(int i = 0; i < objectsInScene.Length; i++)
		{
			Transform transform = objectsInScene[i].transform;
			objectNames.Add(transform.name);
		}
		
		foreach (var name in objectNames.OrderBy(q => q).ToList())
		{
			fileName += "_" + name;
		}
		fileName = camera.transform.position.GetHashCode() + "_" + camera.transform.rotation.GetHashCode() + "_" + fileName.GetHashCode() + fileEnding;
		directOccluders = new Dictionary<string, List<Transform>> ();
		directOccluders = loadFile(fileName);
		if (directOccluders.Any())
        {
            Debug.Log("Occlusion Graph from XML loaded. Filename: " + fileName);
            return;
        }

        int currentLayer = 0; 
		GameObject[] allObjects = GameObject.FindGameObjectsWithTag(Main.tagName);
		var objects = allObjects.Length;
		while (objects > 0) {
			allObjects = GameObject.FindGameObjectsWithTag(Main.tagName);
			Layer layer = new Layer (currentLayer);
			List<Transform> allObjectsList = new List<Transform>();
			Dictionary<string,int> indices = new Dictionary<string, int>();
			for (int i = 0; i < allObjects.Length; i++) {
				allObjectsList.Add (allObjects [i].transform);
				indices.Add(allObjects[i].transform.name, i);
			}
			int[] visibleCount = new int[allObjects.Length];	//Saves the numbers of visible hits for every object
			int[] occlusionCount = new int[allObjects.Length];		//Saves the number of occlusions for every object
			int index = -1;
			for (int i = 0; i < Screen.width; i++) {
				for (int j = 0; j < Screen.height; j++) {
					Ray ray = camera.ScreenPointToRay (new Vector2(i,j));
					RaycastHit[] hits = Physics.RaycastAll (ray);
					float min = Mathf.Infinity;
					int minIndex = -1;
					for (int hitIndex = 0; hitIndex < hits.Length; hitIndex++)
					{
						index = indices[hits[hitIndex].transform.name];
						occlusionCount[index]++;
						var magCamToHit = (hits[hitIndex].point - camera.transform.position).magnitude;
						if (magCamToHit < min) {
							min = magCamToHit;
							minIndex = hitIndex;
						}
					}
					if(minIndex == -1)
						continue;
					index = indices[hits[minIndex].transform.name];
					occlusionCount[index]--;
					visibleCount[index]++;
				}
			}

            float maxVisibilityRatio = 0;
            int maxObj = 0; 
			//Check layer
			for (int i = 0; i < allObjectsList.Count; i++) {
				float visibilityRatio = (float)(visibleCount[i]) / (float)((visibleCount[i] + occlusionCount[i]));
                if(visibilityRatio >= maxVisibilityRatio){
                    maxVisibilityRatio = visibilityRatio;
                    maxObj = i;
                }
                if (visibilityRatio >= layerVisibilityRATIO) {
					layer.addTransform (allObjectsList[i]);
					allObjectsList [i].gameObject.SetActive (false);
					objects--;
				}
			}
            if (layer.getTransformList().Count == 0)
            {
                layer.addTransform(allObjectsList[maxObj]);
                allObjectsList[maxObj].gameObject.SetActive(false);
                objects--;
            }
            layers.Add (layer);
		//	Debug.Log (layer.toString());
			currentLayer++;
		}
		for (int i = 0; i < objectsInScene.Length; i++) {
			objectsInScene [i].SetActive (true);
		}
		
	}

	/// <summary>
	/// Compute the occlusion graph for a viewpoint. An occlusion graph saves the direct occluder for each element in the scene.
	/// </summary>
	/// <param name="camera">The camera specifies the viewpoint for the occlusion graph.</param>
	public void computeGraph(Camera camera){
		if (directOccluders.Any())
			return;
		GameObject[] objectsInScene = GameObject.FindGameObjectsWithTag(Main.tagName);
		List<Transform> objectsInSceneList = new List<Transform>();
		string fileName = "";
		List<string> objectNames = new List<string>();
		for(int i = 0; i < objectsInScene.Length; i++)
		{
			Transform transform = objectsInScene[i].transform;
			objectsInSceneList.Add (transform);
			objectNames.Add(transform.name);
		}
		foreach (var name in objectNames.OrderBy(q => q).ToList())
		{
			fileName += "_" + name;
		}
		fileName = camera.transform.position.GetHashCode() + "_" + camera.transform.rotation.GetHashCode() + "_" + fileName.GetHashCode() + fileEnding;
		Debug.Log ("-----Compute Occlusion Graph------");
		for (int layerIndex = layers.Count -1 ; layerIndex >= 0; layerIndex--) {
			//Debug.Log ("Start with Layer nr. " + layers[layerIndex].getNumber());
			Layer layer = layers [layerIndex];
			List<Transform> objectsOfLayer = layer.getTransformList ();
			foreach (Transform element in objectsOfLayer) {
			//	Debug.Log ("Start with element: " + element.name);
				layer.setAllInvisible ();
				element.gameObject.SetActive (true);
				int numberCollision = 0;
				int[] occlusionCount = new int[objectsInScene.Length];		//Saves the number of occlusions for every object
				for (int i = 0; i < Screen.width; i++) {
					for (int j = 0; j < Screen.height; j++) {
						Ray ray = camera.ScreenPointToRay (new Vector2(i,j));
						RaycastHit[] hits = Physics.RaycastAll (ray);
						bool elementHit = false;
						float max = -1;
						Transform directOccluder = null;
						for (int hitI = 0; hitI < hits.Length; hitI++)
						{
							var magCamToHit = (hits[hitI].point - camera.transform.position).magnitude;
							if (hits [hitI].transform.name.Equals (element.name)) {
								elementHit = true;
							} else if(magCamToHit > max)
							{
								max = magCamToHit;
								directOccluder = hits [hitI].transform;
							}
						}
						if (elementHit) {
							numberCollision++;
							if(directOccluder != null){
								int indexOccluder = objectsInSceneList.IndexOf (directOccluder);
								occlusionCount [indexOccluder] = occlusionCount [indexOccluder] + 1;
							}
						}
					}
				}
				List<Transform> children = new List<Transform> ();
				for (int i = 0; i < objectsInSceneList.Count; i++) {
					float occlusionRatio= (float)(occlusionCount[i]) / (float)((numberCollision));
					if (occlusionRatio >= graphOcclusionRATIO) {
						//Debug.Log ("DirectOccluder: " + objectsInSceneList[i].name);
						children.Add (objectsInSceneList[i]);
					}
				}
				directOccluders.Add (element.transform.name, children);
				element.gameObject.SetActive (false);
				//layers.ForEach (l => l.setAllVisible ());
			}
		}
		objectsInSceneList.ForEach (o => o.gameObject.SetActive (true));
		saveFile(directOccluders, fileName);
	}
	
	/// <summary>
	/// Checks if a file exists.
	/// </summary>
	/// <param name="fileName">The name of the file to check</param>
	/// <returns>true if the file exists, false if it doesn't</returns>
	private bool doesFileExist(string fileName)
	{
		return File.Exists(@fileName);;
	}
	
	/// <summary>
	/// Loads and reads an occlusion xml file
	/// </summary>
	/// <param name="fileName">The name of the file to load</param>
	/// <returns>The loaded Dictionary</returns>
	private Dictionary<string, List<Transform>> loadFile(string fileName)
	{
		if(!doesFileExist(fileName))
			return new Dictionary<string, List<Transform>> ();
		Debug.Log ("--------Load XML file-----------");
		var serializer = new XmlSerializer(typeof(List<KeyValuePair<string,string>>));
		var stream = new FileStream(fileName, FileMode.Open);
		var directOccludersNames = serializer.Deserialize(stream) as List<KeyValuePair<string,string>>;
		stream.Close();
		var directOccluders = new Dictionary<string,List<Transform>>();
		var keys = new List<string>();
		foreach (var valuePair in directOccludersNames)
		{
			var key = valuePair.Key;
			if (keys.Contains(key))
			{
				var occluder = directOccluders[key];
				keys.Add(key);
				occluder.Add(GameObject.Find(valuePair.Value).transform);
				directOccluders[key] = occluder;
			}
			else
			{
				var occluder = new List<Transform>();
				keys.Add(key);
				var name = valuePair.Value;
				if(name != null)
					occluder.Add(GameObject.Find(name).transform);
				directOccluders.Add(key, occluder);
			}
		}
		return directOccluders;
	}

	/// <summary>
	/// Saves an occlusion xml file
	/// </summary>
	/// <param name="directOccluders">The Dictionary to save</param>
	/// <param name="fileName">The name of the file to save</param>
	private void saveFile(Dictionary<string,List<Transform>> directOccluders, string fileName)
	{
		Debug.Log ("--------Save XML file-----------");
		var directOccludersNames = new List<KeyValuePair<string,string>>();
		foreach(var key in directOccluders.Keys)
		{
			var occluders = directOccluders[key];
			foreach (var oclluder in occluders)
			{
				KeyValuePair<string,string> valuePair = new KeyValuePair<string, string>(key,oclluder.name);
				directOccludersNames.Add(valuePair);
			}
			if (occluders.Count() == 0)
			{
				KeyValuePair<string,string> valuePair = new KeyValuePair<string, string>(key,null);
				directOccludersNames.Add(valuePair);
			}
		}
		var serializer = new XmlSerializer(typeof(List<KeyValuePair<string,string>>));
		var stream = new FileStream(fileName, FileMode.Create);
		serializer.Serialize(stream, directOccludersNames);
		stream.Close();
	}

	public List<Transform> getDirectOccluders(string name){
		return directOccluders [name];
	}


}

/// <summary>
/// The layer class saves a list of geometries that are associated to the layer. The layer with number 0 is the layer with the geometries that are nearest to the camera.
/// </summary>
public class Layer{
	private int number; // 0 is nearest to camera
	private List<Transform> geoList;

	/// <summary>
	/// The constructor of the class Layer
	/// </summary>
	/// <param name="number">The number of the layer</param>
	public Layer(int number){
		this.number = number;
		this.geoList = new List<Transform> ();
	}

	/// <summary>
	/// Adds a transform to the layer
	/// </summary>
	/// <param name="trans">The transform to add</param>
	public void addTransform(Transform trans){
		geoList.Add(trans);
	}
	
	/// <summary>
	/// The toString method of the Layer class
	/// </summary>
	/// <returns>The class described in one string</returns>
	public string toString(){
		string debugInfo = "Layer " + number + " contains geos: ";
		for (int i = 0; i < geoList.Count; i++) {
			debugInfo += geoList[i].name+"; ";
		}
		return debugInfo;
	}

	/// <summary>
	/// Checks if the Layer contains a geometry
	/// </summary>
	/// <param name="refTrans">The transform to check</param>
	/// <returns></returns>
	public bool containsGeo(Transform refTrans){
		return geoList.Any (t => t.name.Equals(refTrans.name));
	}

	/// <summary>
	/// Sets all objects in this layer to invisible
	/// </summary>
	public void setAllInvisible(){
		geoList.ForEach (t => t.gameObject.SetActive (false));
	}

	
	/// <summary>
	/// Sets all objects in this layer to visible
	/// </summary>
	public void setAllVisible(){
		geoList.ForEach (t => t.gameObject.SetActive (true));
	}

	/// <summary>
	/// Teturns the list of added transforms
	/// </summary>
	/// <returns>The list of transforms</returns>
	public List<Transform> getTransformList(){
		return geoList;
	}

	/// <summary>
	/// The number of the layer
	/// </summary>
	/// <returns>The number given on construction</returns>
	public int getNumber(){
		return number;
	}

}